using System;
using System.Data.SqlTypes;
using System.Dynamic;
using System.Net;
using System.Net.Sockets;
using System.Text;
using EvaPosSCOSrv;
using EvaPosSrvResp;
using GatewaySCO;
using Serilog;
using Serilog.Events;

namespace Servidor.TcpGateway;

public class TcpGateway
{
    readonly static bool _isDebug = Log.IsEnabled(LogEventLevel.Debug);
    public int GatewayPort { get; private set; } = 6690;
    public string PosbcServer { get; private set; } = "127.0.0.1";
    public int PosbcPort { get; private set; } = 6697;
    private MessageProcessor _processor = null;

    public TcpGateway(int gatewayPort, string posbcServer, int posbcPort)
    {
        GatewayPort = gatewayPort;
        PosbcServer = posbcServer;
        PosbcPort = posbcPort;
        GatewayPort = gatewayPort;
    }

    public async Task Start()
    {
        try
        {
            // Id proceso. Compilación difiere según .net usado.
            int idProceso = -1;
#if NETFRAMEWORK
                idProceso = Process.GetCurrentProcess().Id;
#elif (NETSTANDARD || NET5_0_OR_GREATER)
            idProceso = Environment.ProcessId;
#endif
            Log.Information("\n" +
                            " _      _  _    __ ___               _   _   __  _   _ \n" +
                            "/  |_| |_ / __ /__  | \\    / \\_/ __ |_) / \\ (_  |_) /  \n" +
                            "\\_ | | |_ \\_   \\_|  |  \\/\\/   |     |   \\_/ __) |_) \\_ \n");
            Log.Information("Configuración del servidor Gateway TCP:");
            Log.Information("\tDirección del servidor Gateway: {ServerAddress}", IPAddress.Any);
            Log.Information("\tPuerto del servidor: {ServerPort}", GatewayPort);
            Log.Information("\tProtocolo: {Protocol}", ProtocolType.Tcp);
            Log.Information("\tVersión framework: {version}", Environment.Version);
            Log.Information("\tId proceso: {id}", idProceso);

            _processor = new MessageProcessor();

            TcpListener gatewayListener = new TcpListener(IPAddress.Any, GatewayPort);
            gatewayListener.Start();

            while (true)
            {
                TcpClient checClient = await gatewayListener.AcceptTcpClientAsync();
                IPEndPoint checEndPoint = (IPEndPoint)checClient.Client.RemoteEndPoint;
                Log.Information("CHEC conectado.");
                Log.Information("\tDirección CHEC: {ChecAddress}", checEndPoint.Address);
                Log.Information("\tPuerto CHEC: {ChecPort}", checEndPoint.Port);
                _ = Task.Run(() => HandleChecClientAsync(checClient));
            }
        }
        catch (Exception ex)
        {
            Log.Error(ex, "Error en el Gateway");
        }
        finally
        {
            Log.CloseAndFlush();
        }
    }

    private async Task HandleChecClientAsync(TcpClient checClient)
    {
        using (checClient)
        using (NetworkStream checStream = checClient.GetStream())
        using (TcpClient posbcClient = new TcpClient())
        {
            try
            {
                await posbcClient.ConnectAsync(PosbcServer, PosbcPort);
                IPEndPoint posbcEndPoint = (IPEndPoint)posbcClient.Client.RemoteEndPoint;
                Log.Information("Conectado a POSBC.");
                Log.Information("\tDirección POSBC: {PosbcAddress}", posbcEndPoint.Address);
                Log.Information("\tPuerto POSBC: {PosbcPort}", posbcEndPoint.Port);
                using (NetworkStream posbcStream = posbcClient.GetStream())
                {
                    Task receiveFromPosbcTask = ReceiveFromPosbcAndSendToChecAsync(posbcStream, posbcClient, checStream, checClient);

                    while (true)
                    {
                        // Leer mensaje desde CHEC
                        byte[] lengthBuffer = new byte[4];
                        int bytesRead = await ReadExactlyAsync(checStream, lengthBuffer, 4);
                        //int bytesRead = await checStream.ReadAsync(lengthBuffer, 0, 4);
                        if (bytesRead == 0)
                        {
                            Log.Debug("CHEC cabecera bytesRead = 0");
                            continue;
                        } // Si no se lee nada, continuar esperando

                        int messageLength = Util.LongitudCodificada(lengthBuffer);
                        Log.Debug("Mensaje de CHEC longitud {longitud}", messageLength);
                        byte[] messageBuffer = new byte[messageLength];
                        // Leer el mensaje completo basado en la longitud esperada
                        bytesRead = await ReadExactlyAsync(checStream, messageBuffer, messageLength);
                        if (bytesRead == 0)
                        {
                            Log.Debug("CHEC cuerpo bytesRead = 0");
                            continue;
                        } // Si no se lee nada, continuar esperando

                        // Intercepta mensajes, procesa, y eventualmente generar modificaciones o
                        // mensajes al POSBC
                        _processor.MessageFromCHEC(messageBuffer, messageLength);
                        // TODO: en este punto, los resultados del _processor eventualmente 
                        // se debería integrar a los mensajes de salida.

                        // Remitir mensaje a POSBC
                        byte[] bytesMsj = Util.ConcatenaArreglosBytes(lengthBuffer, messageBuffer);
                        await posbcStream.WriteAsync(bytesMsj, 0, bytesMsj.Length);

                        Log.Information($"Mensaje remitido a POSBC: {Encoding.UTF8.GetString(messageBuffer)}");
                    }
                    await receiveFromPosbcTask;
                }
            }
            catch (Exception ex)
            {
                //Log.Error(ex, "Error manejando la conexión CHEC");
                Log.Warning("Conexión CHEC cerrada.");
            }
        }
    }

    private async Task ReceiveFromPosbcAndSendToChecAsync(NetworkStream posbcStream, TcpClient posbcClient, NetworkStream checStream, TcpClient checClient)
    {
        bool continuar = true;
        while (continuar)
        {
            try
            {
                // Leer longitud del mensaje desde POSBC
                byte[] lengthBuffer = new byte[4];
                int bytesRead = await ReadExactlyAsync(posbcStream, lengthBuffer, 4);
                if (bytesRead == 0)
                {
                    Log.Debug("POSBC cabecera bytesRead = 0");
                    continue;
                } // Si no se lee nada, continuar esperando

                int messageLength = Util.LongitudCodificada(lengthBuffer);
                Log.Debug("POSBC bytes cabecera leidos {bytes}, longitud en cabecera {longitud}", bytesRead, messageLength);
                byte[] messageBuffer = new byte[messageLength];
                bytesRead = await ReadExactlyAsync(posbcStream, messageBuffer, messageLength);
                Log.Debug("POSBC bytes cuerpo leidos {bytes}", bytesRead);
                if (bytesRead == 0)
                {
                    Log.Debug("POSBC cuerpo bytesRead = 0");
                    continue;
                } // Si no se lee nada, continuar esperando
                if (bytesRead == 0) continue; // Si no se lee nada, continuar esperando

                // Intercepta mensajes y eventualmente generar modificaciones.
                // Retorna false si encuenta mensaje de terminación, se sale del ciclo y cierra sesión.
                continuar = _processor.MessageFromPOSBCToCHEC(messageBuffer, messageLength);

                // TODO: en este punto, los resultados del _processor eventualmente 
                // se debería integrar a los mensajes de salida.

                // Remitir mensaje a CHEC
                byte[] bytesMsj = Util.ConcatenaArreglosBytes(lengthBuffer, messageBuffer);
                await checStream.WriteAsync(bytesMsj, 0, bytesMsj.Length);

                Log.Information($"Mensaje remitido a CHEC: {Encoding.UTF8.GetString(messageBuffer)}");

                if (continuar == false)
                {
                    Log.Debug("Mensaje de terminación, continuar es FALSE");
                    posbcStream.Close();
                    posbcClient.Close();

                    checStream.Close();
                    checClient.Close();
                }
            }
            catch (Exception ex)
            {
                Log.Error(ex, "Error recibiendo desde POSBC o enviando a CHEC");
                break;
            }
        }
        Log.Information("Terminación conexión");
    }

    private async Task<int> ReadExactlyAsync(NetworkStream stream, byte[] buffer, int length)
    {
        int totalBytesRead = 0;
        while (totalBytesRead < length)
        {
            int bytesRead = await stream.ReadAsync(buffer, totalBytesRead, length - totalBytesRead);
            if (bytesRead == 0)
            {
                return 0; // La conexión se ha cerrado
            }
            totalBytesRead += bytesRead;
        }
        return totalBytesRead;
    }


}
